/**
 * State Space Module
 * @author James D Bloom (UI) & Clare Clark (Maths)
 *
 * This module will have its code cleaned up (soon)
 * and more comments will be added so help
 * anyone who are tring to understand how is works
 */

package pipe.modules.stateSpace;

import pipe.dataLayer.*;
import pipe.modules.Module;
import pipe.gui.*;
import pipe.modules.invariantAnalysis.*;
import javax.swing.*;
import javax.swing.text.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import javax.swing.border.*;
import java.util.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.event.MouseInputAdapter;

public class StateSpace implements Module {

  // Main Frame
  private static final String MODULE_NAME = "State Space Analysis";
  private JFrame guiFrame = new JFrame("State Space Analysis Module");

  // Dimensions
  private final Toolkit mainToolKit = guiFrame.getToolkit(); // Get the window toolkit
  private final Dimension windowSize = mainToolKit.getScreenSize(); // Get screen size
  private final Point center = GraphicsEnvironment.getLocalGraphicsEnvironment().getCenterPoint(); // Get screen center
  private final int windowWidth = (windowSize.width*55)/100; // Window width in pixels
  private final int windowHeight = (windowSize.height*70)/100; // Window height in pixels
  private final int windowTopLeftCornerX = center.x-windowWidth/2; // Set window position
  private final int windowTopLeftCornerY = center.y-windowHeight/2; // Distance from top left corner of screen

  // Panes
  private JEditorPane codeEditorPane = new JEditorPane();
  private JScrollPane codeScrollPane = new JScrollPane(codeEditorPane);
  private JEditorPane consoleEditorPane = new JEditorPane();
  private JScrollPane consoleScrollPane = new JScrollPane(consoleEditorPane);
  private JScrollPane projectScrollPane = new JScrollPane(codeEditorPane);
  private JScrollPane classScrollPane = new JScrollPane(codeEditorPane);
  private JPanel guiResultsPanel = null;

  // Panels
  private JPanel mainPanel = new JPanel();
  private GridBagLayout gridbag = new GridBagLayout();
  private GridBagConstraints gridbagConstraints = new GridBagConstraints();
  private JScrollPane deadlockScrollPane = null;
  private JPanel deadlockPanel = null;

  // Text Fields
  private final int TEXT_FIELD_SIZE = 15;
  private JTextField sourceTextField = null;
  private JTextField deadlockTextField = new JTextField(35);


  // Other
  private JButton sourceButton = new JButton();
  private JLabel sourceLabel = new JLabel();
  private final Font defaultLabel = null;
  private final Font smallDefaultLabel = null;
  private DataLayer sourcedataLayer = null;
  private DataLayer refdataLayer = null; // Storeage (maintain reference) for input dataLayer parameter;
  private File outputHTML = null;
  private FileWriter outputWriter = null;
  private Insets spacing = new Insets((windowHeight*15)/1000,(windowWidth*15)/1000,(windowHeight*15)/1000,(windowWidth*15)/1000);

  // Checkboxes
  private JCheckBox sourceCheckBox = new JCheckBox();

  // Actions
  final static int NO_ACTION = -1;
  final static int CURRENT_NET = 0;
  final static int SOURCE_BROWSE = 1;
  final static int COMPARISON_BROWSE = 2;
  final static int CLASSIFY = 3;

  // States
  boolean bounded = false;
  boolean safe = false;
  boolean deadlock = false;
  int[] deadlockPath = null;

  public void run(DataLayer pnmlData) {
//    warnUser(pnmlData.getURI(), guiFrame);
    StateSpace stateSpace = new StateSpace(pnmlData);
  }

  public String getName() {
    return MODULE_NAME;
  }

  public StateSpace(DataLayer pnmlData) {
    sourcedataLayer = pnmlData;
    sourceTextField = new JTextField(sourcedataLayer.getURI(), TEXT_FIELD_SIZE);
    deadlockTextField.setFont(new Font("Arial", Font.BOLD, 14));
    try {
      initGuiFrame();
      } catch (java.lang.Exception e) {}
  }

  public StateSpace() {
    sourceTextField = new JTextField(TEXT_FIELD_SIZE);
  }

  private void initGuiFrame() throws Exception {
    guiFrame.setBounds(windowTopLeftCornerX, windowTopLeftCornerY, windowWidth, windowHeight);
    guiFrame.setResizable(false);
    guiFrame.addMouseMotionListener(new MouseHandler());
    Container contentGuiPane = guiFrame.getContentPane();
    contentGuiPane.setLayout(new BorderLayout());
    JPanel guiPetriNetPanel = new JPanel();
    guiPetriNetPanel.addMouseMotionListener(new MouseHandler());
    guiPetriNetPanel.setToolTipText("Click on the \"Analyse\" button to perform a state-space analysis of the Petri-Net");
    setUpPetriNetPanel(guiPetriNetPanel);
    guiPetriNetPanel.setVisible(true);
    contentGuiPane.add(guiPetriNetPanel, BorderLayout.NORTH);

    // Setup Classify Button
    JButton compareButton = new JButton("Analyse");
    compareButton.addActionListener(new HandleControl(CLASSIFY));
    compareButton.setFont(new java.awt.Font("Dialog", 1, 18));
    compareButton.setToolTipText("Analyse the Petri-Net");
    JPanel comparePane = new JPanel();
    comparePane.addMouseMotionListener(new MouseHandler());
    comparePane.setLayout(gridbag);
    gridbagConstraints.insets = spacing;
    gridbag.setConstraints(compareButton, gridbagConstraints);
    comparePane.add(compareButton);
    contentGuiPane.add(comparePane, BorderLayout.SOUTH);
    setUpResultsPanel();
    guiFrame.setVisible(true);
  }

  private void setUpPetriNetPanel(JPanel panel){
    sourceCheckBox.addActionListener(new HandleControl(CURRENT_NET));
    sourceButton.addActionListener(new HandleControl(SOURCE_BROWSE));

    JPanel sourcePanel = new JPanel();
    sourcePanel.addMouseMotionListener(new MouseHandler());
    sourcePanel.setLayout(new FlowLayout());

    // Setup Source Box
    sourceCheckBox.setText("use current net");
    sourceCheckBox.setSelected(true);

    sourceButton.setText("Browse");
    sourceButton.setEnabled(false);
    sourceLabel.setText("Source-Net URI:");
    sourceLabel.setEnabled(false);
    sourceTextField.setEnabled(false);
    sourcePanel.setLayout(new FlowLayout());

    sourcePanel.add(sourceCheckBox);
    sourcePanel.add(sourceLabel);
    sourcePanel.add(sourceTextField);
    sourcePanel.add(sourceButton);

    sourcePanel.setBorder(new TitledBorder(BorderFactory.createEtchedBorder(Color.white,new Color(165, 163, 151)),"Source Petri-Net", TitledBorder.DEFAULT_JUSTIFICATION, TitledBorder.DEFAULT_POSITION, defaultLabel, Color.BLACK));

    // Add to panel
    panel.setLayout(gridbag);
    gridbagConstraints.insets = spacing;
    gridbag.setConstraints(sourcePanel, gridbagConstraints);
    panel.add(sourcePanel);

    panel.setToolTipText("Specify the location of a Petri-Net");
  }

  private void setUpDeadlockPanel(JPanel panel){
/*
    JEditorPane deadlockPath = new JEditorPane();

    try {
      outputHTML = new File("DeadlockPath.html");
      outputHTML.createNewFile();
      outputWriter = new FileWriter(outputHTML);
      outputHTML.deleteOnExit();
      outputWriter.write("<html><body>"); // INCORRECT HTML STANDARD
      outputWriter.write("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;");
      outputWriter.write("</body></html>");
      outputWriter.close();
      deadlockPath.setPage("file:" + outputHTML.getAbsolutePath());
      deadlockPath.setEditable(false);
      outputHTML.delete();
      } catch (IOException dioe) { }
*/

    deadlockPanel = new JPanel();
    deadlockPanel.addMouseMotionListener(new MouseHandler());
//      deadlockPanel.setMaximumSize(new Dimension(15,20));
//      deadlockScrollPane = new JScrollPane(deadlockPath);
    deadlockScrollPane = new JScrollPane(deadlockTextField);
//      deadlockPanel.setLayout(new FlowLayout());
    deadlockPanel.add(deadlockScrollPane);
    deadlockPanel.setBorder(new TitledBorder(BorderFactory.createEtchedBorder(Color.white,new Color(165, 163, 151)),"Shortest Path to Deadlock", TitledBorder.DEFAULT_JUSTIFICATION, TitledBorder.DEFAULT_POSITION, defaultLabel, Color.BLACK));
    panel.add(deadlockPanel);
    panel.setToolTipText("Shortest Path to Deadlock");
    deadlockTextField.setEnabled(true);
    deadlockTextField.setEditable(false);
    deadlockTextField.setBackground(Color.WHITE);
    deadlockScrollPane.setEnabled(false);
    deadlockPanel.setEnabled(false);
  }


  private void setUpResultsPanel() throws IOException {
    guiResultsPanel = new JPanel();
    guiResultsPanel.addMouseMotionListener(new MouseHandler());

    Color background = guiResultsPanel.getBackground();
    background.getRed();
    background.getBlue();
    background.getGreen();

    int red = background.getRed();
    int blue = background.getBlue();
    int green = background.getGreen();

    outputHTML = new File("StateSpaceResults.html");

    outputHTML.createNewFile();
    outputWriter = new FileWriter(outputHTML);

    outputHTML.deleteOnExit();


    outputWriter.write("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n");
    outputWriter.write("<META http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");
    outputWriter.write("<title></title>\n</head>\n<body bgcolor=\"rgb(" + red + "," + green + "," + blue + ")\">\n");
    outputWriter.write("<table align=\"center\">\n");
    outputWriter.write("<tr>\n");
    outputWriter.write("<th nowrap=\"nowrap\" colspan=\"2\"><img alt=\"Results\" src=\"Images/Results.png\"></th></tr>\n");
    outputWriter.write("<tr>\n");
    outputWriter.write("<th nowrap=\"nowrap\"><img alt=\"Bounded\" src=\"Images/Bounded.png\"></th>\n");
    outputWriter.write("<th nowrap=\"nowrap\"><img alt=\"Unknown\" src=\"Images/Unknown.png\"></th></tr>\n");
    outputWriter.write("<tr>\n");
    outputWriter.write("<th nowrap=\"nowrap\"><img alt=\"Safe\" src=\"Images/Safe.png\"></th>\n");
    outputWriter.write("<th nowrap=\"nowrap\"><img alt=\"Unknown\" src=\"Images/Unknown.png\"></th></tr>\n");
    outputWriter.write("<tr>\n");
    outputWriter.write("<th nowrap=\"nowrap\"><img alt=\"Deadlock\" src=\"Images/Deadlock.png\"></th>\n");
    outputWriter.write("<th nowrap=\"nowrap\"><img alt=\"Unknown\" src=\"Images/Unknown.png\"></th></tr>\n");
    outputWriter.write("</body>\n</html>\n");
    outputWriter.close();
    displayResults();
    outputHTML.delete();

  }

  private void warnUser(String message, JFrame frame) {
    JOptionPane.showMessageDialog(frame, message, "Petri-Net Classification Module", JOptionPane.CLOSED_OPTION);
  }

  private void classifyPetriNets() throws IOException {
    guiResultsPanel.setVisible(false);
    Color background = guiResultsPanel.getBackground();
    background.getRed();
    background.getBlue();
    background.getGreen();

    int red = background.getRed();
    int blue = background.getBlue();
    int green = background.getGreen();

    outputHTML = new File("StateSpaceResults.html");
    outputHTML.delete();
    outputHTML.createNewFile();
    outputWriter = new FileWriter(outputHTML);

    outputHTML.deleteOnExit();


    outputWriter.write("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">");
    outputWriter.write("<META http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">");

    outputWriter.write("</head>\n<body bgcolor=\"rgb(" + red + "," + green + "," + blue + ")\">\n");
    outputWriter.write("<table align=\"center\">");
    outputWriter.write("<tr>");
    outputWriter.write("<th nowrap=\"nowrap\"  colspan=\"2\"><img src=\"Images/Results.png\"></th>");
    outputWriter.write("</tr><tr>");
    outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/Bounded.png\"></th>");
    if(bounded)
      outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/Yes.png\"></th>");
    else
      outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/No.png\"></th>");
    outputWriter.write("</tr><tr>");
    outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/Safe.png\"></th>");
    if(safe)
      outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/Yes.png\"></th>");
    else
      outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/No.png\"></th>");
    outputWriter.write("</tr><tr>");
    outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/Deadlock.png\"></th>");
    if(deadlock)
      outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/Yes.png\"></th>");
    else
      outputWriter.write("<th nowrap=\"nowrap\" ><img src=\"Images/No.png\"></th>");
    outputWriter.write("</tr>");
    outputWriter.write("</table>");
    outputWriter.write("</body>\n</html>");
    outputWriter.close();
    displayResults();
    guiResultsPanel.setVisible(true);
    outputHTML.delete();
  }

  private void displayResults() throws IOException {
    Color background = guiResultsPanel.getBackground();
    background.getRed();
    background.getBlue();
    background.getGreen();

    int red = background.getRed();
    int blue = background.getBlue();
    int green = background.getGreen();

    JEditorPane editorPane = new JEditorPane();
    editorPane.setBackground(background);
    editorPane.setPage("file:" + outputHTML.getAbsolutePath());
    editorPane.setEditable(false);
    JScrollPane scrollPane = new JScrollPane(editorPane);

    int w = windowWidth*1/2;
    int h = windowHeight*3/4;
    int x = center.x-w/2;
    int y = center.y-h/2;
    editorPane.setBounds(x, y, w, h);
    Dimension dimension = new Dimension(w, h);
    editorPane.setMinimumSize(dimension);
    scrollPane.setMinimumSize(dimension);
    guiResultsPanel.setMinimumSize(dimension);
    guiResultsPanel = new JPanel();
    guiResultsPanel.addMouseMotionListener(new MouseHandler());
    guiResultsPanel.add(editorPane);
    guiResultsPanel.setVisible(true);
    guiResultsPanel.setBorder(BorderFactory.createEtchedBorder(Color.white,new Color(165, 163, 151)));
    JPanel guiResultsOuterPanel = new JPanel();
    guiResultsOuterPanel.addMouseMotionListener(new MouseHandler());
    JPanel guiResultsOuterOuterPanel = new JPanel();
    guiResultsOuterOuterPanel.addMouseMotionListener(new MouseHandler());
    guiResultsOuterPanel.setLayout(new BoxLayout(guiResultsOuterPanel, BoxLayout.Y_AXIS));
    gridbagConstraints.insets = spacing;
    gridbag.setConstraints(guiResultsPanel, gridbagConstraints);
    guiResultsOuterPanel.add(new JPanel());
    guiResultsOuterPanel.add(guiResultsPanel);
    guiResultsOuterPanel.add(new JPanel());
    guiResultsOuterPanel.add(new JPanel());
    setUpDeadlockPanel(guiResultsOuterPanel);
    guiResultsOuterOuterPanel.add(guiResultsOuterPanel);
    guiFrame.getContentPane().add(guiResultsOuterOuterPanel, BorderLayout.CENTER);
  }

  class MouseHandler extends MouseInputAdapter {

    public void mouseMoved(MouseEvent e) {
      guiResultsPanel.setVisible(false);
      guiResultsPanel.setVisible(true);
      deadlockPanel.setVisible(false);
      deadlockPanel.setVisible(true);
    }
  }

  // Class defining a handler for a control button
  class HandleControl implements ActionListener {
    private int buttonID;

    // Constructor
    public HandleControl(int buttonID) {
      this.buttonID = buttonID; // Store the button ID
    }

    private String openFile(File file, String title, JTextField textField) {
      int fileChooserReturnVal = JFileChooser.ERROR_OPTION;
      JFileChooser fileChooser = null;
      if(sourcedataLayer.getURI() != null)
        fileChooser = new JFileChooser(sourcedataLayer.getURI());
      else
        fileChooser = new JFileChooser();

      String explain = "Please select " + title;

      fileChooser.setToolTipText(explain);
      fileChooser.setName(explain);
      fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
      fileChooser.setDialogTitle(explain);
      fileChooser.setFileSelectionMode(fileChooser.FILES_ONLY);
      ExtensionFilter netFilter = new ExtensionFilter(".xml","Petri-Net files (*.xml)");

      fileChooser.setAcceptAllFileFilterUsed(false); // Remove 'all files'filter
      fileChooser.addChoosableFileFilter(netFilter); // add Filter
      fileChooser.setFileFilter(netFilter); // and select it
      fileChooser.rescanCurrentDirectory();

      if(file != null)
        fileChooser.setSelectedFile(file);

      fileChooserReturnVal = fileChooser.showOpenDialog(guiFrame);

      if (fileChooserReturnVal == JFileChooser.APPROVE_OPTION) {

        file = fileChooser.getSelectedFile();

        if (file.getName().endsWith(".xml")) {

          try {

            if(file != null)
              sourcedataLayer = new DataLayer(file);

          } catch (java.io.IOException ex) {
            warnUser("java.io.IOException " + ex + " File Access Error with " + title, guiFrame);
          } catch (org.xml.sax.SAXException ex) {
            warnUser("org.xml.sax.SAXException " + ex + " File Access Error with " + title, guiFrame);
          } catch (javax.xml.transform.TransformerException  ex) {
            warnUser("javax.xml.transform.TransformerException " + ex + " File Access Error with " + title, guiFrame);
          } catch (javax.xml.parsers.ParserConfigurationException  ex) {
            warnUser("javax.xml.parsers.ParserConfigurationException " + ex + " File Access Error with " + title, guiFrame);
          }

        } else {
          int returnVal = JOptionPane.showConfirmDialog(guiFrame, "Incorrect file type selected do you want to try again?", "Please select a Petri-Net File (*.xml)", JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
          if(returnVal == JOptionPane.YES_OPTION) {
            textField.setText(null);
            file = null;
            return openFile(file, title, textField);
          }
        }

      }

      if(file == null) return null;

      return file.getName();
    }

    // Handle button click
    public void actionPerformed(ActionEvent e) {
      File sourceFile = null;

      switch(buttonID) {
        case NO_ACTION:
          break;
        case CURRENT_NET:
          guiResultsPanel.setVisible(false);
          guiResultsPanel.setVisible(true);
          if(sourceCheckBox.isSelected()) {
            sourceButton.setEnabled(false);
            sourceLabel.setEnabled(false);
            sourceTextField.setEnabled(false);
            sourcedataLayer = refdataLayer;
          } else {
            sourceButton.setEnabled(true);
            sourceLabel.setEnabled(true);
            sourceTextField.setEnabled(true);
            refdataLayer = sourcedataLayer;
          }
          sourceTextField.setText(sourcedataLayer.getURI());
          break;
        case SOURCE_BROWSE:
          guiResultsPanel.setVisible(false);
          guiResultsPanel.removeAll();
          sourceTextField.setText(openFile(sourceFile,  "the Source Petri-Net file", sourceTextField));
          try {
            setUpResultsPanel();
            } catch (IOException ioe) { warnUser("Internal File Error", guiFrame); }
            break;
          case CLASSIFY:
            bounded = false;
            safe = false;
            deadlock = false;

            // Call Clare's function
            int[] markup = sourcedataLayer.getCurrentMarkupMatrix();
            int[][] forwards = sourcedataLayer.getFowardsIncidenceMatrix();
            int[][] backwards = sourcedataLayer.getBackwardsIncidenceMatrix();
//            Transition[] transitions = sourcedataLayer.getTransitions();
            myTree tree = null;

            if(markup != null && forwards !=null && backwards != null) {
              tree = new myTree(markup, new IntMatrix(forwards), new IntMatrix(backwards));
              bounded = !tree.Found_An_Omega;
              safe = !tree.more_Than_One_Token;
              deadlock = tree.no_Enabled_Transitions;
              if (tree.tooBig)
              {
                warnUser("There are more than 10,000 nodes in the state space tree. Tree expansion aborted. Results will be incomplete.", guiFrame);
              }
            }
            if(deadlock) {
              int i = 0;
              String jamie = new String("  ");
              for (i=0; i<tree.pathToDeadlock.length; i++)
              {
                if(sourcedataLayer.getTransition(tree.pathToDeadlock[i]-1) != null) {
                  if(sourcedataLayer.getTransition(tree.pathToDeadlock[i]-1).getName() != null) {
                    jamie += sourcedataLayer.getTransition(tree.pathToDeadlock[i]-1).getName() + " ";
                  }
                }
              }
              deadlockTextField.setText(jamie);
            } else {
              deadlockTextField.setText("");
            }
            guiResultsPanel.setVisible(false);
            try {
              classifyPetriNets();
              } catch (IOException ioe) { warnUser("Internal File Error", guiFrame); }

              guiResultsPanel.setVisible(true);
              break;
      }
    }


    class myTree {

      public boolean Found_An_Omega = false;  // bounded
      public boolean more_Than_One_Token = false;  // safe
      public boolean no_Enabled_Transitions = false;  // deadlock

      myNode root;      //root of the tree
      boolean[] transitions;    //each bool corresponds to a transition
      boolean[] nodes;            //each bool corresponds to a node
      int nodeCount = 0;    //Total number of nodes in tree
      IntMatrix CPlus;
      IntMatrix CMinus;                 //incidence matrices of petri net
      int number_transitions;   //number of transitions in net
      int number_places;    //number of places in the net
      myNode[] EndNodes;    //The end nodes (leaves) in the tree
      int[] pathToDeadlock;   //Gives transitions to deadlock
      boolean tooBig = false; // Set if the tree gets too large

      //Tree Constructor
      public myTree(int[] tree_root, IntMatrix plus, IntMatrix minus) {
        CPlus = plus;
        CMinus = minus;

//        System.out.print("\nNew Tree Created \n" );

        //Find number transitions in net from incidence matrix dimensions
        number_transitions = CMinus.getColumnDimension();

        //Find number of places in net from incidence matrix dimensions
        number_places = CMinus.getRowDimension();

        //Create root of tree by calling Node constructor
        root = new myNode(tree_root, this, 1);

        //Call expansion function on root of tree
        root.RecursiveExpansion();

      }
    }


    class myNode {

      boolean[] trans_array;     //array of transitions
      myNode parent;             //parent node
      myNode[] children;         //child nodes
      myTree tree;               //tree that contains the node
      int[] Markup;      //The marking of the node
      myNode previousInstance;   //same node found in tree
      int depth;        // The depth this node is in the tree

      //Attributes used for assessing whether a node has occured before
      boolean Repeated_Node = false;

      /*-------------------------------------------------------------------
        Function: void myNode(int[] marking_array, myTree tree)

        Node constructor called by the tree object

      -------------------------------------------------------------------*/

      public myNode(int [] marking_array, myTree atree, int treeDepth) {
        int i;             //counter for 'for printing loop'

        //Node now knows its tree in order to use its attributes
        tree = atree;

        depth = treeDepth;

        //Update the count of nodes
        tree.nodeCount++;

        //Set up marking for this node
        Markup = marking_array;  //created in myTree

//        System.out.print("Node "+tree.nodeCount+" Created \n" );

        //Assign the root nodes parents (this constructor only called by tree)
        parent = tree.root; //Just get root to have itself as parent

        //Create an array of transitions for each node
        trans_array = new boolean[tree.number_transitions];

        //Create an array of nodes for each node (for children)
        //Number of children limited by total number of transitions
        children = new myNode[tree.number_transitions];

        //Initialise trans_array
        for(i=0; i< tree.number_transitions; i++){
          trans_array[i] = false;
        }

//        for(i=0; i< Markup.length; i++) {
//          System.out.print(this.Markup[i]);
//          System.out.print(" ");
//        }



      }


      /*-------------------------------------------------------------------
        Function: void myNode(int[] marking_array, myNode parent_node, myTree atree)

        The overloaded Node constructor called by a node object

      -------------------------------------------------------------------*/

      public myNode(int [] marking_array, myNode parent_node, myTree atree, int treeDepth) {
        int i;             //counter for 'for printing loop'

        //Node now knows its tree in order to use its attributes
        tree = atree;

        depth = treeDepth;

        //Update the count of nodes
        tree.nodeCount++;

        //Set up marking for this node
        Markup = marking_array;  //created in myTree

        //Assign the nodes parents
        parent = parent_node;

//        System.out.print("Node "+tree.nodeCount+" Created \n" );

        //Create an array of transitions for each node
        trans_array = new boolean[tree.number_transitions];

        //Create an array of nodes for each node (for children)
        //Number of children limited to total number of transitions.
        children = new myNode[tree.number_transitions];

        //Initialise trans_array
        for(i=0; i< tree.number_transitions; i++){
          trans_array[i] = false;
        }

//        for(i=0; i< Markup.length; i++) {
//          System.out.print(this.Markup[i]);
//          System.out.print(" ");
//        }

      }
    /*---------------------------------------------------------------------
       Function: boolean TransitionEnabled(int transIndex)

      Tests to see if a particular transition is enabled by consulting C-
      and the current markup contained in the tree.
      transIndex = usual petri net description of a transition by an integer

      ---------------------------------------------------------------------*/

      public  boolean TransitionEnabled(int transIndex) {
        int count;        //index for 'for loop'
        int CMinusValue;

        for(count=0; count < tree.number_places; count++)
        {

          CMinusValue = (tree.CMinus).get(count,(transIndex-1));

          if(Markup[count] < CMinusValue && Markup[count]!=-1)
          {
            //There is a place where marking is less than required in CMinus
            return false;
          }
        }

        //All places satisfy the marking criteria
        return true;

      }

    /*---------------------------------------------------------------------
       Function: int[] fire(int trans)

       Produces a new markup vector to simulate the firing of a transition.
       Destroys the number of tokens shown in CMinus for a given place and
       transition, and creates the number of tokens shown in CPlus.

       TransIndex refers to the actual transition number ie starting at 1.
      ---------------------------------------------------------------------*/

      public int[] fire(int transIndex) {
        int count;                     //index for 'for loop'
        int CMinusValue;         //Value from C- matrix
        int CPlusValue;          //Value from C+ matrix

        //Create marking array to return
        int[] marking = new int[tree.number_places];

        for(count = 0; count < tree.number_places; count++)
        {
          CMinusValue = (tree.CMinus).get(count, (transIndex - 1));
          CPlusValue = (tree.CPlus).get(count, (transIndex - 1));

          if (Markup[count]!=-1)
            marking[count] = Markup[count] - CMinusValue + CPlusValue;
          else
            marking[count] = Markup[count];
        }

        //Return this new marking to RecursiveExpansion function
        return marking;
      }

      /*----------------------------------------------------------------------
         Function: void RecursiveExpansion()

        Undertakes a recursive expansion of the tree
        Called on root node from within the tree constructor.

      -----------------------------------------------------------------------*/

      public void RecursiveExpansion() {
        int transIndex;                              //Index to count transitions
        int[] new_markup;                            //markup used to create new node
        int i;                     //index for loops
        boolean A_Transition_Is_Enabled = false;     //To check for deadlock
        boolean allOmegas;

        //For each transition
        for(transIndex = 1; transIndex <= tree.number_transitions; transIndex++)
        {
          //If it is enabled as described by CMinus and Markup
          if(TransitionEnabled(transIndex) == true)
          {
          //Set trans_array of to true for this index
          //index 0 refers to transition 1 here.
          trans_array[transIndex - 1] = true;

//          System.out.println("\n Transition " +transIndex+ " Enabled \n" );
          A_Transition_Is_Enabled = true;

          //Fire transition to produce new markup vector
          new_markup = fire(transIndex);

//          System.out.println("\n Old Markup is");
//          for(i=0; i< Markup.length; i++){
//            System.out.print(Markup[i]);
//            System.out.print(" ");
//        }
//        System.out.print("\n");

          //Check for safeness. If any of places have > 1 token set variable.
          for(i=0; i< new_markup.length; i++){
            if(new_markup[i] > 1 || new_markup[i]==-1) {
              tree.more_Than_One_Token = true;
            }

            //Print new markup
//            System.out.print(new_markup[i]);
          }

          //Create a new node using the new markup vector and attach it
          //to the current node as a child.

          children[transIndex - 1] = new myNode(new_markup, this, tree, depth+1);

          /* Now need to (a) check if any omegas (represented by -1) need to be
          inserted in the markup vector of the new node, and (b) if the resulting
          markup vector has appeared anywhere else in the tree. We must do (a) before
          (b) as we need to know if the new node contains any omegas to be able to
          compare it with existing nodes. */

          allOmegas = children[transIndex - 1].InsertOmegas();

          //check if the resulting markup vector has occured anywhere else in the tree

          Repeated_Node = (tree.root).FindMarkup(children[transIndex-1]);

          if (tree.nodeCount>=10000 && !tree.tooBig)
          {
            tree.tooBig = true;
          }

          if(!Repeated_Node && !allOmegas)
          {
            children[transIndex - 1].RecursiveExpansion();
          }
        }

        }

        if(!A_Transition_Is_Enabled)
        {
//          System.out.println("No transition enabled");
          if(!tree.no_Enabled_Transitions || tree.pathToDeadlock.length<depth-1)
          {
            RecordDeadlockPath();
            tree.no_Enabled_Transitions = true;
          }
          else
          {
//			System.out.println("Deadlocked node found, but path is not shorter than current path.");
          }
        }
        else
        {
//			System.out.println("Transitions enabled.");
        }
      }

      /*----------------------------------------------------------------------
         Function: void RecordDeadlockPath()

         If there is a deadlock, calculates the path
      -----------------------------------------------------------------------*/

      public void RecordDeadlockPath() {
        myNode currentNode;      //The current node we're considering
        int[] path;              //returned showing path to deadlock
        int pos;                 //position in path array
        int i;

        //Set up array to return
        tree.pathToDeadlock = new int[depth-1];
        pos = depth-2; // Start filling in at the end of the array

        currentNode = this;

        //For each ancestor node until root
        while(currentNode != tree.root)
        {
          // Work out which transition we followed to get to to currentNode
          loop: for(i = 1; i <= tree.number_transitions; i++)
          {
            if ( currentNode.parent.trans_array[i-1]
                 && currentNode.parent.children[i-1]==currentNode )
                 {
                   // That's the one!
            break loop;
          }
          }

          tree.pathToDeadlock[pos] = i;
          pos--;
          //Update current node to look at an earlier ancestor
          currentNode = currentNode.parent;
        }

//          System.out.print("Path to deadlock is: ");
//          for (i=0; i<depth-1; i++)
//          {
//            System.out.print(tree.pathToDeadlock[i] + " ");
//          }
//          System.out.print("\n");
      }


      /*----------------------------------------------------------------------
         Function: void InsertOmegas()

         Checks if any omegas need to be inserted in the places of a given node.
         Omegas (shown by -1 here) represent unbounded places and are therefore
         important when testing whether a petri net is bounded. This function
         checks each of the ancestors of a given node.
         Returns true iff all places now contain an omega.

      -----------------------------------------------------------------------*/

      public boolean InsertOmegas() {

        //Attributes used for assessing boundedness of the net
        boolean All_Elements_Greater_Or_Equal = true;
        boolean [] Element_Is_Strictly_Greater;
        myNode ancestor_node;       //one of the ancestors of this node
        int i;                      //counter for the 'for loop'

        //Set up array used for checking boundedness
        Element_Is_Strictly_Greater = new boolean[tree.number_places];

        //Initialise this array to false
        for(i=0; i< tree.number_places; i++){
          Element_Is_Strictly_Greater[i] = false;
        }

        ancestor_node = this;

        //For each ancestor node until root
        while(ancestor_node != tree.root)
        {
          //Update ancestor node to look at an earlier ancestor
          ancestor_node = ancestor_node.parent;

          All_Elements_Greater_Or_Equal = true;

          //compare markup of this node with that of each ancestor node
          //for each place in the markup
          loop: for(i=0; i < tree.number_places; i++)
          {
            if (Markup[i]!=-1)
            {
              //If M(p) for new node less than M(p) for ancestor
              if(Markup[i] < ancestor_node.Markup[i])
              {
            All_Elements_Greater_Or_Equal = false;
            break loop;
          }

          //If M(p) for new node strictly greater than M(p) for ancestor
          Element_Is_Strictly_Greater[i] = (Markup[i] > ancestor_node.Markup[i]);
          }
          }

          //Now assess the information obtained for this node
          if(All_Elements_Greater_Or_Equal == true)
          {
            //for each place p in the markup of this node
            for(i=0; i < tree.number_places; i++)
            {
              if(Markup[i]!=-1 && Element_Is_Strictly_Greater[i])
              {
                //Set M(p) in this new markup to be omega
              Markup[i] = -1;
//              System.out.println("\n Omega added");
//              for(i=0; i< Markup.length; i++) {
//                System.out.print(this.Markup[i]);
//                System.out.print(" ");
//              }
//                System.out.print("\n");

              //Set variable in tree to true - net unbound
              tree.Found_An_Omega = true;
            }
            }
          }
        }

        for(i=0; i< tree.number_places; i++)
        {
          if (Markup[i]!=-1)
          {
          return false;
        }
        }

        return true; // All places have omegas

      }

      /*----------------------------------------------------------------------
         Function: boolean FindMarkup(int[] mark)

         Checks if the markup has occured previously in the tree. This means
         previously during the creation of the tree, and not just on the current
         branch as a direct ancestor of the node. Function updates arrays of
         end nodes stored in the tree class for the liveness tests. Explore
         the tree by starting with the root node and investigating the children.

      -----------------------------------------------------------------------*/

      public boolean FindMarkup(myNode n)
      {
        if(n==this)
        {
          return false;
        }

        int i;  //counter for 'for loop'

        if(MarkupCompare(n.Markup))
        {

          n.previousInstance = this;
//          System.out.println("\nFound duplicate markup!\n");
          return true;
        }

        for(i=0; i< tree.number_transitions; i++)
        {
          if(trans_array[i])
          {
          if(children[i].FindMarkup(n))
          {
          return true;
        }
        }
        }
        return false;
      }

      /*----------------------------------------------------------------------
         Function: boolean MarkupCompare(int[] check)

         Takes two integer arrays (Markups) and compares the values.
         Returns a boolean showing the results of the comparison.

      -----------------------------------------------------------------------*/

      public boolean MarkupCompare(int[] check) {
        int i;     //counter for 'for loop'

        //Comparison only possible on same length arrays
        if(this.Markup.length == check.length)
        {
          for(i=0; i < Markup.length; i++)
          {
            if(this.Markup[i] != check[i])
            {
            return false;
          }
          }
          return true;
        }

        else return false;
      }

    }
  }

}